@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2015-2017 Fanout, Inc.
+ * Copyright (C) 2015-2019 Fanout, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -1029,13 +1029,13 @@ private:
 
 			if(!responseSent)
 			{
-				// apply filters of all channels to content
+				// apply ProxyContent filters of all channels
 				QStringList allFilters;
 				foreach(const Instruct::Channel &c, instruct.channels)
 				{
 					foreach(const QString &filter, c.filters)
 					{
-						if(!allFilters.contains(filter))
+						if((Filter::targets(filter) & Filter::ProxyContent) && !allFilters.contains(filter))
 							allFilters += filter;
 					}
 				}
@@ -1146,15 +1146,13 @@ private:
 				{
 					if(!c.prevId.isNull())
 					{
-						QString name = channelPrefix + c.name;
-
 						if(!rp.haveInspectInfo)
 						{
 							rp.haveInspectInfo = true;
 							rp.inspectInfo.doProxy = true;
 						}
 
-						rp.inspectInfo.lastIds.insert(name.toUtf8(), c.prevId.toUtf8());
+						rp.inspectInfo.lastIds.insert(c.name.toUtf8(), c.prevId.toUtf8());
 					}
 				}
 
@@ -1505,23 +1503,31 @@ public:
 			log_info("in pull: %s", qPrintable(config.pushInSpec));
 		}
 
-		if(!config.pushInSubSpec.isEmpty())
+		if(!config.pushInSubSpecs.isEmpty())
 		{
 			inSubSock = new QZmq::Socket(QZmq::Socket::Sub, this);
 			inSubSock->setSendHwm(SUB_SNDHWM);
 			inSubSock->setShutdownWaitTime(0);
 
 			QString errorMessage;
-			if(!ZUtil::setupSocket(inSubSock, config.pushInSubSpec, true, config.ipcFileMode, &errorMessage))
+			if(!ZUtil::setupSocket(inSubSock, config.pushInSubSpecs, !config.pushInSubConnect, config.ipcFileMode, &errorMessage))
 			{
 					log_error("%s", qPrintable(errorMessage));
 					return false;
 			}
 
+			if(config.pushInSubConnect)
+			{
+				// some sane TCP keep-alive settings
+				// idle=30, cnt=6, intvl=5
+				inSubSock->setTcpKeepAliveEnabled(true);
+				inSubSock->setTcpKeepAliveParameters(30, 6, 5);
+			}
+
 			inSubValve = new QZmq::Valve(inSubSock, this);
 			connect(inSubValve, &QZmq::Valve::readyRead, this, &Private::inSub_readyRead);
 
-			log_info("in sub: %s", qPrintable(config.pushInSubSpec));
+			log_info("in sub: %s", qPrintable(config.pushInSubSpecs.join(", ")));
 		}
 
 		if(!config.retryOutSpec.isEmpty())
@@ -1789,6 +1795,23 @@ private:
 		{
 			WsSession *s = qobject_cast<WsSession*>(target);
 
+			if(f.haveContentFilters)
+			{
+				// ensure content filters match
+				QStringList contentFilters;
+				foreach(const QString &f, s->channelFilters[item.channel])
+				{
+					if(Filter::targets(f) & Filter::MessageContent)
+						contentFilters += f;
+				}
+				if(contentFilters != f.contentFilters)
+				{
+					QString errorMessage = QString("content filter mismatch: subscription=%1 message=%2").arg(contentFilters.join(","), f.contentFilters.join(","));
+					log_debug("%s", qPrintable(errorMessage));
+					return;
+				}
+			}
+
 			Filter::Context fc;
 			fc.subscriptionMeta = s->meta;
 			fc.publishMeta = item.meta;
@@ -2207,8 +2230,8 @@ private slots:
 				out["command"] = config.commandSpec.toUtf8();
 			if(!config.pushInSpec.isEmpty())
 				out["publish-pull"] = config.pushInSpec.toUtf8();
-			if(!config.pushInSubSpec.isEmpty())
-				out["publish-sub"] = config.pushInSubSpec.toUtf8();
+			if(!config.pushInSubSpecs.isEmpty() && !config.pushInSubConnect)
+				out["publish-sub"] = config.pushInSubSpecs[0].toUtf8();
 			req->respond(out);
 			delete req;
 		}
@@ -2231,6 +2254,56 @@ private slots:
 		}
 	}
 
+	QVariant parseJsonOrTnetstring(const QByteArray &message, bool *ok = 0, QString *errorMessage = 0) {
+		QVariant data;
+		bool ok_;
+		if(message.length() > 0 && message[0] == 'J') {
+			QJsonParseError e;
+			QJsonDocument doc = QJsonDocument::fromJson(message.mid(1), &e);
+			if(e.error != QJsonParseError::NoError)
+			{
+				if(errorMessage)
+					*errorMessage = QString("received message with invalid format (json parse failed)");
+				if(ok)
+					*ok = false;
+				return data;
+			}
+
+			if(doc.isObject())
+			{
+				data = doc.object().toVariantMap();
+			}
+			else
+			{
+				if(errorMessage)
+					*errorMessage = QString("received message with invalid format (not a valid json object)");
+				if(ok)
+					*ok = false;
+				return data;
+			}
+		}
+		else
+		{
+			int offset = 0;
+			if(message.length() > 0 && message[0] == 'T') {
+				offset = 1;
+			}
+
+			data = TnetString::toVariant(message, offset, &ok_);
+			if(!ok_)
+			{
+				if(errorMessage)
+					*errorMessage = QString("received message with invalid format (tnetstring parse failed)");
+				if(ok)
+					*ok = false;
+				return data;
+			}
+		}
+		if(ok)
+			*ok = true;
+		return data;
+	}
+
 	void inPull_readyRead(const QList<QByteArray> &message)
 	{
 		if(message.count() != 1)
@@ -2240,17 +2313,17 @@ private slots:
 		}
 
 		bool ok;
-		QVariant data = TnetString::toVariant(message[0], 0, &ok);
+		QString errorMessage;
+		QVariant data = parseJsonOrTnetstring(message[0], &ok, &errorMessage);
 		if(!ok)
 		{
-			log_warning("IN pull: received message with invalid format (tnetstring parse failed), skipping");
+			log_warning("IN pull: %s, skipping", qPrintable(errorMessage));
 			return;
 		}
 
 		if(log_outputLevel() >= LOG_LEVEL_DEBUG)
 			log_debug("IN pull: %s", qPrintable(TnetString::variantToString(data, -1)));
 
-		QString errorMessage;
 		PublishItem item = PublishItem::fromVariant(data, QString(), &ok, &errorMessage);
 		if(!ok)
 		{
@@ -2270,10 +2343,10 @@ private slots:
 		}
 
 		bool ok;
-		QVariant data = TnetString::toVariant(message[1], 0, &ok);
-		if(!ok)
-		{
-			log_warning("IN sub: received message with invalid format (tnetstring parse failed), skipping");
+		QString errorMessage;
+		QVariant data = parseJsonOrTnetstring(message[1], &ok, &errorMessage);
+		if(!ok) {
+			log_warning("IN sub: %s, skipping", qPrintable(errorMessage));
 			return;
 		}
 
@@ -2282,7 +2355,6 @@ private slots:
 		if(log_outputLevel() >= LOG_LEVEL_DEBUG)
 			log_debug("IN sub: channel=%s %s", qPrintable(channel), qPrintable(TnetString::variantToString(data, -1)));
 
-		QString errorMessage;
 		PublishItem item = PublishItem::fromVariant(data, channel, &ok, &errorMessage);
 		if(!ok)
 		{
@@ -2486,6 +2558,11 @@ private slots:
 						s->keepAliveType = contentType;
 						s->keepAliveMessage = cm.content;
 
+						if(cm.keepAliveMode == "interval")
+							i.keepAliveMode = "interval";
+						else
+							i.keepAliveMode = "idle";
+
 						i.timeout = (cm.timeout > 0 ? cm.timeout : DEFAULT_WS_KEEPALIVE_TIMEOUT);
 					}
 					else
