@@ -393,7 +393,7 @@ public:
 			return;
 		}
 
-		if(!ensureDir(logDir))
+		if(!args.mergeOutput && !ensureDir(logDir))
 		{
 			log_error("failed to create directory: %s", qPrintable(logDir));
 			emit q->quit(1);
@@ -464,14 +464,14 @@ public:
 				m2shBin = settings.value("runner/m2sh_bin").toString();
 
 			QString certsDir = QDir(configDir).filePath("certs");
-			if(!Mongrel2Service::generateConfigFile(m2shBin, QDir(libDir).filePath("mongrel2.conf.template"), runDir, logDir, ipcPrefix, filePrefix, certsDir, clientBufferSize, interfaces))
+			if(!Mongrel2Service::generateConfigFile(m2shBin, QDir(libDir).filePath("mongrel2.conf.template"), runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, certsDir, clientBufferSize, interfaces, logLevel))
 			{
 				emit q->quit(1);
 				return;
 			}
 
 			foreach(const Mongrel2Service::Interface &i, interfaces)
-				services += new Mongrel2Service(m2Bin, QDir(runDir).filePath(QString("%1mongrel2.sqlite").arg(filePrefix)), "default_" + QString::number(i.port), runDir, logDir, filePrefix, i.port, i.ssl, this);
+				services += new Mongrel2Service(m2Bin, QDir(runDir).filePath(QString("%1mongrel2.sqlite").arg(filePrefix)), "default_" + QString::number(i.port), runDir, !args.mergeOutput ? logDir : QString(), filePrefix, i.port, i.ssl, logLevel, this);
 		}
 
 		if(serviceNames.contains("m2adapter"))
@@ -480,7 +480,7 @@ public:
 			foreach(const Mongrel2Service::Interface &i, interfaces)
 				ports += i.port;
 
-			services += new M2AdapterService(m2aBin, QDir(libDir).filePath("m2adapter.conf.template"), runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, logLevel >= 3, ports, this);
+			services += new M2AdapterService(m2aBin, QDir(libDir).filePath("m2adapter.conf.template"), runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, logLevel, ports, this);
 		}
 
 		bool quietCheck = false;
@@ -491,17 +491,17 @@ public:
 			if(settings.contains("runner/zurl_bin"))
 				zurlBin = settings.value("runner/zurl_bin").toString();
 
-			services += new ZurlService(zurlBin, QDir(libDir).filePath("zurl.conf.template"), runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, logLevel >= 3, this);
+			services += new ZurlService(zurlBin, QDir(libDir).filePath("zurl.conf.template"), runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, logLevel, this);
 
 			// when zurl is managed by pushpin, log updates checks as debug level
 			quietCheck = true;
 		}
 
 		if(serviceNames.contains("pushpin-proxy"))
-			services += new PushpinProxyService(proxyBin, configFile, runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, logLevel >= 3, args.routeLines, quietCheck, this);
+			services += new PushpinProxyService(proxyBin, configFile, runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, logLevel, args.routeLines, quietCheck, this);
 
 		if(serviceNames.contains("pushpin-handler"))
-			services += new PushpinHandlerService(handlerBin, configFile, runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, portOffset, logLevel >= 3, this);
+			services += new PushpinHandlerService(handlerBin, configFile, runDir, !args.mergeOutput ? logDir : QString(), ipcPrefix, filePrefix, portOffset, logLevel, this);
 
 		foreach(Service *s, services)
 		{
@@ -598,8 +598,10 @@ private slots:
 	{
 		Service *s = (Service *)sender();
 
-		QString out = tryInsertPrefix(line, '[' + s->name() + "] ");
-		log_raw(qPrintable(out));
+		QString out = tryInsertPrefix(s->formatLogLine(line), '[' + s->name() + "] ");
+		if(!out.isEmpty()) {
+			log_raw(qPrintable(out));
+		}
 	}
 
 	void service_error(const QString &error)
